/** @file camogm_align.c
 * @brief Provides frame alignment functions use for recording to block device.
 * @copyright Copyright (C) 2017 Elphel, Inc.
 *
 * @par <b>License</b>
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/** @brief This define is needed to use lseek64 and should be set before includes */
#define _LARGEFILE64_SOURCE
#define ALIGN_INTERFRAME // New alignment mode - add zeros between frames (TIFF also), not inside JPEG with app15 (JPEG only)
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/uio.h>

#include "camogm_align.h"

#ifndef ALIGN_INTERFRAME
static unsigned char app15[ALIGNMENT_SIZE] = {0xff, 0xef};
#endif

//static inline size_t get_size_from(const struct iovec *vects, int index, size_t offset, int all); // will not be used
static inline size_t get_size_from_paged(camogm_state *state, int index, int all);
static inline size_t align_bytes_num(size_t data_len, size_t align_len);
static inline void vectcpy(struct iovec *dest, void *src, size_t len);
static inline void vectset(struct iovec *dest, int src, size_t len);
static inline void vectshrink(struct iovec *vec, size_t len);
static inline void vectshrinkhead(struct iovec *vec, size_t len);
static inline unsigned char *vectrpos(struct iovec *vec, size_t offset);
//static void dev_dbg(const char *prefix, const char *format, ...);
static void remap_vectors(camogm_state *state, struct iovec *chunks);
static size_t get_blocks_num(struct iovec *sgl, size_t n_elem);
/* debug functions */
static int check_chunks(struct iovec *vects, int page);
static void dev_dbg(const char *prefix, const char *format, ...);


/** Replace debug function with the same name from driver code with debug macro to reduce changes */
static void dev_dbg(const char *prefix, const char *format, ...)
{
	va_list args;

	va_start(args, format);
	D6(vfprintf(debug_file, format, args));
	va_end(args);
}

/** Copy @e len bytes from buffer pointed by @e src vector to buffer pointed by @e dest vector */
static inline void vectcpy(struct iovec *dest, void *src, size_t len)
{
	unsigned char *d = (unsigned char *)dest->iov_base;

	memcpy(d + dest->iov_len, src, len);
	dest->iov_len += len;
}

/** Set @e len bytes with @e src to buffer pointed by @e dest vector */
static inline void vectset(struct iovec *dest,  int src, size_t len)
{
	unsigned char *d = (unsigned char *)dest->iov_base;

	memset(d + dest->iov_len, src, len);
	dest->iov_len += len;
}



/** Shrink vector length by @len bytes */
static inline void vectshrink(struct iovec *vec, size_t len)
{
	if (vec->iov_len >= len) {
		vec->iov_len -= len;
	}
}

/** Shrink vector length by @len bytes from the head */
static inline void vectshrinkhead(struct iovec *vec, size_t len)
{
	if (vec->iov_len >= len) {
		vec->iov_len -= len;
		unsigned char * p = (unsigned char * ) vec->iov_base; // += len;
		vec->iov_base = p + len;
	}
}


/** This helper function is used to position a pointer @e offset bytes from the end
 * of a buffer. */
static inline unsigned char *vectrpos(struct iovec *vec, size_t offset)
{
	return (unsigned char *)vec->iov_base + (vec->iov_len - offset);
}

#if 0
/** Calculate the size of current frame in bytes starting from vector and offset given */
static inline size_t get_size_from(const struct iovec *vects, int index, size_t offset, int all) // offset, index - always 0
{
	int i;
//	int page_offs = FILE_CHUNKS_NUM + page;
	size_t total = 0;

	if (index >= MAX_DATA_CHUNKS || offset > vects[index].iov_len) {
		return 0;
	}
	for (i = index; i < MAX_DATA_CHUNKS; i++) {
		if (i == CHUNK_REM && all == EXCLUDE_REM)
			// remainder should not be processed
			continue;
		if (i == index)
			total += vects[i].iov_len - offset;
		else
			total += vects[i].iov_len;
	}

	return total;
}
#endif

//camogm_state *state
/** Calculate the size of current frame in bytes starting from vector and offset given */
static inline size_t get_size_from_paged(camogm_state *state, int index, int all)
{
	int i;
	struct iovec *chunks = state->writer_params.data_chunks;
	struct iovec *cbuff = &chunks[CHUNK_COMMON];
	int page_offs = (state->writer_params.chunk_page_prep % FILE_CHUNKS_PAGES) * FILE_CHUNKS_NUM;
	size_t total = 0;
	for (i = page_offs + index; i < (FILE_CHUNKS_NUM + page_offs); i++) {
		total += chunks[i].iov_len;
	}
	if (all) {
		total +=  state->writer_params.was_CHUNK_REM.iov_len;
	}
	return total;

}




/** Return the number of bytes needed to be inserted to align @e data_len to @e align_len boundary */
static inline size_t align_bytes_num(size_t data_len, size_t align_len)
{
	size_t rem = data_len % align_len;
	if (rem == 0)
		return 0;
	else
		return align_len - rem;
}

/** Remap vectors pointing to various buffers with frame data to vectors used during frame alignment */
static void remap_vectors(camogm_state *state, struct iovec *chunks)
{
	int chunk_index = 1;
    int is_tiff = state->this_frame_params[state->port_num].color == COLORMODE_RAW;
    int chunk_offset = (state->writer_params.chunk_page_prep % FILE_CHUNKS_PAGES) * FILE_CHUNKS_NUM;
	int paged_CHUNK_LEADER =  chunk_offset + CHUNK_LEADER;
	int paged_CHUNK_EXIF =    chunk_offset + CHUNK_EXIF;
	int paged_CHUNK_HEADER =  chunk_offset + CHUNK_HEADER;
	int paged_CHUNK_COMMON =  chunk_offset + CHUNK_COMMON;
	int paged_CHUNK_DATA_0 =  chunk_offset + CHUNK_DATA_0;
	int paged_CHUNK_DATA_1 =  chunk_offset + CHUNK_DATA_1;
	int paged_CHUNK_TRAILER = chunk_offset + CHUNK_TRAILER;
	int paged_CHUNK_ALIGN =   chunk_offset + CHUNK_ALIGN;
	if (state->exif > 0) {
		if (is_tiff) { // Tiff
			chunks[paged_CHUNK_EXIF].iov_base =   (void *)state->packetchunks[chunk_index].chunk;
			chunks[paged_CHUNK_EXIF].iov_len =            state->packetchunks[chunk_index++].bytes;
			// set length of unused chunks to zero, and pointer to any valid pointer
			chunks[paged_CHUNK_LEADER].iov_base = chunks[paged_CHUNK_EXIF].iov_base;
			chunks[paged_CHUNK_LEADER].iov_len = 0;
			chunks[paged_CHUNK_HEADER].iov_base = chunks[paged_CHUNK_EXIF].iov_base;
			chunks[paged_CHUNK_HEADER].iov_len = 0;
		} else { // JPEG/JP4
			chunks[paged_CHUNK_LEADER].iov_base = (void *)state->packetchunks[chunk_index].chunk;
			chunks[paged_CHUNK_LEADER].iov_len =          state->packetchunks[chunk_index++].bytes;
			chunks[paged_CHUNK_EXIF].iov_base =   (void *)state->packetchunks[chunk_index].chunk;
			chunks[paged_CHUNK_EXIF].iov_len =            state->packetchunks[chunk_index++].bytes;
			chunks[paged_CHUNK_HEADER].iov_base = (void *)state->packetchunks[chunk_index].chunk;
			chunks[paged_CHUNK_HEADER].iov_len =          state->packetchunks[chunk_index++].bytes;
		}
	} else {
		chunks[paged_CHUNK_LEADER].iov_base = (void *)state->packetchunks[chunk_index].chunk;
		chunks[paged_CHUNK_LEADER].iov_len  =          JPEG_MARKER_LEN;
		chunks[paged_CHUNK_HEADER].iov_base = (void *)(state->packetchunks[chunk_index].chunk + JPEG_MARKER_LEN);
		chunks[paged_CHUNK_HEADER].iov_len  =          state->packetchunks[chunk_index++].bytes - JPEG_MARKER_LEN;
		// set to zero length and valid pointer
		chunks[paged_CHUNK_EXIF].iov_base   = chunks[paged_CHUNK_HEADER].iov_base;
		chunks[paged_CHUNK_EXIF].iov_len    = 0;
	}
	chunks[paged_CHUNK_DATA_0].iov_base = (void *)state->packetchunks[chunk_index].chunk;
	chunks[paged_CHUNK_DATA_0].iov_len = state->packetchunks[chunk_index++].bytes;
	if (state->writer_params.segments == 2) {
		chunks[paged_CHUNK_DATA_1].iov_base = (void *)state->packetchunks[chunk_index].chunk;
		chunks[paged_CHUNK_DATA_1].iov_len = state->packetchunks[chunk_index++].bytes;
	} else{
		chunks[paged_CHUNK_DATA_1].iov_base = chunks[paged_CHUNK_DATA_0].iov_base;
		chunks[paged_CHUNK_DATA_1].iov_len =  0;
	}
//	chunks[CHUNK_TRAILER].iov_base = (void *)state->packetchunks[chunk_index].chunk;
	if (is_tiff) { // Tiff
		chunks[paged_CHUNK_TRAILER].iov_base = (void *)state->packetchunks[chunk_index-1].chunk; // last used chunk (valid)
		chunks[paged_CHUNK_TRAILER].iov_len = 0;
	} else {
		chunks[paged_CHUNK_TRAILER].iov_base = (void *)state->packetchunks[chunk_index].chunk;
		chunks[paged_CHUNK_TRAILER].iov_len = state->packetchunks[chunk_index].bytes;
	}


	/* some data may be left from previous frame, copy it to special buffer */
	/*
	if (chunks[CHUNK_REM].iov_len != 0) {
		vectcpy(&state->writer_params.prev_rem_vect, chunks[CHUNK_REM].iov_base, chunks[CHUNK_REM].iov_len);
		vectshrink(chunk_rem, chunks[CHUNK_REM].iov_len);
	}
    */
	if (state->writer_params.was_CHUNK_REM.iov_len != 0) {
		vectcpy(&state->writer_params.prev_rem_vect, state->writer_params.was_CHUNK_REM.iov_base, state->writer_params.was_CHUNK_REM.iov_len);
		vectshrink(&(state->writer_params.was_CHUNK_REM), state->writer_params.was_CHUNK_REM.iov_len); // VERIFY !!! &()
	}
}

/** Debug function, checks frame alignment */
static int check_chunks(struct iovec *vects, int page)
{
	int i;
	int ret = 0;
	size_t sz = 0;
    int chunk_offset =        page * FILE_CHUNKS_NUM;
	for (i = chunk_offset; i < FILE_CHUNKS_NUM + chunk_offset; i++) {
//		if (i != CHUNK_REM) {
			sz += vects[i].iov_len;
			if ((vects[i].iov_len % ALIGNMENT_SIZE) != 0) {
				dev_dbg(NULL, "ERROR: unaligned write from slot %d, length %u\n", i, vects[i].iov_len);
				ret = -1;
			}
//		}
		dev_dbg(NULL, "chunk[%02d]: ptr = %p, size = %d\n", i, vects[i].iov_base, vects[i].iov_len);
	}
	if ((sz % PHY_BLOCK_SIZE) != 0) {
		dev_dbg(NULL, "ERROR: total length of the transaction is not aligned to sector boundary, total length %u\n", sz);
		ret = -1;
	} else {
		dev_dbg(NULL, "===== frame is OK =====\n");

	}
	return ret;
}





/** Calculate the number of blocks this frame will occupy. The frame must be aligned to block size */
static size_t get_blocks_num(struct iovec *sgl, size_t n_elem)
{
	int num;
	size_t total = 0;

	for (num = 0; num < n_elem; num++) {
		total += sgl[num].iov_len;
	}

	return total / PHY_BLOCK_SIZE;
}

/** Allocate and initialize buffers for frame alignment */
int init_align_buffers(camogm_state *state)
{
	int i;
//	state->writer_params.data_chunks = (struct iovec *)malloc(MAX_DATA_CHUNKS * sizeof(struct iovec));
	state->writer_params.data_chunks = (struct iovec *)malloc(FILE_CHUNKS_TOTAL * sizeof(struct iovec)); // does not include CHUNK_REM
	if (state->writer_params.data_chunks == NULL) {
		return -1;
	}
	for (i= 0; i < FILE_CHUNKS_PAGES; i++) {
		state->writer_params.common_buffs[i] = (unsigned char *)malloc(COMMON_BUFF_SZ);
		if (state->writer_params.common_buffs[i] == NULL) {
			deinit_align_buffers(state);
			return -1;
		}
	}
	state->writer_params.rem_buff = (unsigned char *)malloc(REM_BUFF_SZ);
	if (state->writer_params.rem_buff == NULL) {
		deinit_align_buffers(state);
		return -1;
	}
	state->writer_params.prev_rem_buff = (unsigned char *)malloc(REM_BUFF_SZ);
	if (state->writer_params.prev_rem_buff == NULL) {
		deinit_align_buffers(state);
		return -1;
	}
	for (i= 0; i < FILE_CHUNKS_PAGES; i++) {
		state->writer_params.data_chunks[FILE_CHUNKS_NUM* i + CHUNK_COMMON].iov_base = (void *)state->writer_params.common_buffs[i];
		state->writer_params.data_chunks[FILE_CHUNKS_NUM* i + CHUNK_COMMON].iov_len = 0;
	}
//was_CHUNK_REM
//	state->writer_params.data_chunks[CHUNK_REM].iov_base = (void *)state->writer_params.rem_buff;
//	state->writer_params.data_chunks[CHUNK_REM].iov_len = 0;
	// this is a single array
	state->writer_params.was_CHUNK_REM.iov_base = (void *)state->writer_params.rem_buff;
	state->writer_params.was_CHUNK_REM.iov_len = 0;

	state->writer_params.prev_rem_vect.iov_base = (void *)state->writer_params.prev_rem_buff;
	state->writer_params.prev_rem_vect.iov_len = 0;

	return 0;
}

/** Delete buffers for frame alignment */
void deinit_align_buffers(camogm_state *state)
{
	struct writer_params *params = &state->writer_params;
    int i;
	if (params->data_chunks) {
		free(params->data_chunks);
		params->data_chunks = NULL;
	}
	for (i= 0; i < FILE_CHUNKS_PAGES; i++) {
		if (params->common_buffs[i]) {
			free(params->common_buffs[i]);
			params->common_buffs[i] = NULL;
		}
	}
	if (params->rem_buff) {
		free(params->rem_buff);
		params->rem_buff = NULL;
	}
	if (params->prev_rem_buff) {
		free(params->prev_rem_buff);
		params->prev_rem_buff = NULL;
	}
}

/** Align current frame to disk sector boundary and each individual buffer to #ALIGNMENT_SIZE boundary */
void align_frame(camogm_state *state)
{
	const char *dev = NULL;
	unsigned char *src;
	size_t len, total_sz, data_len;
	int page = state->writer_params.chunk_page_prep % FILE_CHUNKS_PAGES;
    int chunk_offset =        page * FILE_CHUNKS_NUM;
	int paged_CHUNK_LEADER =  chunk_offset + CHUNK_LEADER;
	int paged_CHUNK_EXIF =    chunk_offset + CHUNK_EXIF;
	int paged_CHUNK_HEADER =  chunk_offset + CHUNK_HEADER;
	int paged_CHUNK_COMMON =  chunk_offset + CHUNK_COMMON;
	int paged_CHUNK_DATA_0 =  chunk_offset + CHUNK_DATA_0;
	int paged_CHUNK_DATA_1 =  chunk_offset + CHUNK_DATA_1;
	int paged_CHUNK_TRAILER = chunk_offset + CHUNK_TRAILER;
	int paged_CHUNK_ALIGN =   chunk_offset + CHUNK_ALIGN;
	struct iovec *chunks =    state->writer_params.data_chunks;
	struct iovec *chunk_rem = &(state->writer_params.was_CHUNK_REM);
	struct iovec *cbuff = &chunks[paged_CHUNK_COMMON];
	struct iovec *rbuff = &state->writer_params.prev_rem_vect;

	remap_vectors(state, chunks); // now multi-page

//	total_sz = get_size_from(chunks, 0, 0, INCLUDE_REM) + rbuff->iov_len;
	total_sz = get_size_from_paged(state, 0, INCLUDE_REM) + rbuff->iov_len;
	D4(fprintf(debug_file, "total_sz = %d, rbuff->iov_len=%d\n", total_sz, (int)(rbuff->iov_len)));
	if (total_sz < PHY_BLOCK_SIZE) {
		/* the frame length is less than sector size, delay this frame */
		if (rbuff->iov_len != 0) {
			/* some data may be left from previous frame */
//			vectcpy(&chunks[CHUNK_REM], rbuff->iov_base, rbuff->iov_len);
			vectcpy(chunk_rem, rbuff->iov_base, rbuff->iov_len);
			//state->writer_params.was_CHUNK_REM
			vectshrink(rbuff, rbuff->iov_len);
			D4(fprintf(debug_file, "copied rbuff->iov_len = %d\n", rbuff->iov_len));
		}
		dev_dbg(dev, "frame size is less than sector size: %u bytes; delay recording\n", total_sz);
//		vectcpy    (&chunks[CHUNK_REM],     chunks[CHUNK_LEADER].iov_base,  chunks[CHUNK_LEADER].iov_len);
//		vectshrink (&chunks[CHUNK_LEADER],  chunks[CHUNK_LEADER].iov_len);
//		vectcpy    (&chunks[CHUNK_REM],     chunks[CHUNK_EXIF].iov_base,    chunks[CHUNK_EXIF].iov_len);
//		vectshrink (&chunks[CHUNK_EXIF],    chunks[CHUNK_EXIF].iov_len);
//		vectcpy    (&chunks[CHUNK_REM],     chunks[CHUNK_HEADER].iov_base,  chunks[CHUNK_HEADER].iov_len);
//		vectshrink (&chunks[CHUNK_HEADER],  chunks[CHUNK_HEADER].iov_len);
//		vectcpy    (&chunks[CHUNK_REM],     chunks[CHUNK_DATA_0].iov_base,  chunks[CHUNK_DATA_0].iov_len);
//		vectshrink (&chunks[CHUNK_DATA_0],  chunks[CHUNK_DATA_0].iov_len);
//		vectcpy    (&chunks[CHUNK_REM],     chunks[CHUNK_DATA_1].iov_base,  chunks[CHUNK_DATA_1].iov_len);
//		vectshrink (&chunks[CHUNK_DATA_1],  chunks[CHUNK_DATA_1].iov_len);
//		vectcpy    (&chunks[CHUNK_REM],     chunks[CHUNK_TRAILER].iov_base, chunks[CHUNK_TRAILER].iov_len);
//		vectshrink (&chunks[CHUNK_TRAILER], chunks[CHUNK_TRAILER].iov_len);
		vectcpy    (chunk_rem,                    chunks[paged_CHUNK_LEADER].iov_base,  chunks[paged_CHUNK_LEADER].iov_len);
		vectshrink (&chunks[paged_CHUNK_LEADER],  chunks[paged_CHUNK_LEADER].iov_len);
		vectcpy    (chunk_rem,                    chunks[paged_CHUNK_EXIF].iov_base,    chunks[paged_CHUNK_EXIF].iov_len);
		vectshrink (&chunks[paged_CHUNK_EXIF],    chunks[paged_CHUNK_EXIF].iov_len);
		vectcpy    (chunk_rem,                    chunks[paged_CHUNK_HEADER].iov_base,  chunks[paged_CHUNK_HEADER].iov_len);
		vectshrink (&chunks[paged_CHUNK_HEADER],  chunks[paged_CHUNK_HEADER].iov_len);
		vectcpy    (chunk_rem,                    chunks[paged_CHUNK_DATA_0].iov_base,  chunks[paged_CHUNK_DATA_0].iov_len);
		vectshrink (&chunks[paged_CHUNK_DATA_0],  chunks[paged_CHUNK_DATA_0].iov_len);
		vectcpy    (chunk_rem,                    chunks[paged_CHUNK_DATA_1].iov_base,  chunks[paged_CHUNK_DATA_1].iov_len);
		vectshrink (&chunks[paged_CHUNK_DATA_1],  chunks[paged_CHUNK_DATA_1].iov_len);
		vectcpy    (chunk_rem,                    chunks[paged_CHUNK_TRAILER].iov_base, chunks[paged_CHUNK_TRAILER].iov_len);
		vectshrink (&chunks[paged_CHUNK_TRAILER], chunks[paged_CHUNK_TRAILER].iov_len);
		return;
	}

	/* copy remainder of previous frame to the beginning of common buffer */
	if (rbuff->iov_len != 0) {
		len = rbuff->iov_len;
		dev_dbg(dev, "copy %u bytes from REM to common buffer\n", len);
		vectcpy(cbuff, rbuff->iov_base, len);
		vectshrink(rbuff, rbuff->iov_len);
	}

	/*Calculate alignment size and add zeros here*/
#ifdef ALIGN_INTERFRAME
	// calculate length after adding CHUNK_LEADER, CHUNK_EXIF and CHUNK_HEADER to the current cbuf
	data_len = cbuff->iov_len + chunks[paged_CHUNK_LEADER].iov_len + chunks[paged_CHUNK_EXIF].iov_len+ chunks[paged_CHUNK_HEADER].iov_len;
	len = align_bytes_num(data_len, ALIGNMENT_SIZE);
	dev_dbg(dev, "total number of stuffing bytes between frames: %u\n", len);
	vectset(cbuff,  0, len);
#endif
	/* copy JPEG marker if present */
	if (chunks[paged_CHUNK_LEADER].iov_len != 0) {
		len = chunks[paged_CHUNK_LEADER].iov_len;
		dev_dbg(dev, "copy %u bytes from LEADER to common buffer\n", len);
		vectcpy(cbuff, chunks[paged_CHUNK_LEADER].iov_base, len);
		vectshrink(&chunks[paged_CHUNK_LEADER], chunks[paged_CHUNK_LEADER].iov_len);
	}
	/* copy Exif if present */
	if (chunks[paged_CHUNK_EXIF].iov_len != 0) {
		len = chunks[paged_CHUNK_EXIF].iov_len;
		dev_dbg(dev, "copy %u bytes from EXIF to common buffer\n", len);
		vectcpy(cbuff, chunks[paged_CHUNK_EXIF].iov_base, len);
		vectshrink(&chunks[paged_CHUNK_EXIF], chunks[paged_CHUNK_EXIF].iov_len);
	}

#ifdef ALIGN_INTERFRAME
	/* copy JPEG JPEG header data excluding leading marker if present */
	if (chunks[paged_CHUNK_HEADER].iov_len != 0) {
		len = chunks[paged_CHUNK_HEADER].iov_len;
		dev_dbg(dev, "copy %u bytes from LEADER to common buffer\n", len);
		vectcpy(cbuff, chunks[paged_CHUNK_HEADER].iov_base, len);
		vectshrink(&chunks[paged_CHUNK_HEADER], chunks[paged_CHUNK_HEADER].iov_len);
	}

#else	// old way with app15
	if (chunks[paged_CHUNK_HEADER].iov_len != 0){ // only if it is not TIFF
		/* align common buffer to ALIGNMENT boundary, APP15 marker should be placed before header data */
		data_len = cbuff->iov_len + chunks[paged_CHUNK_HEADER].iov_len;

		len = align_bytes_num(data_len, ALIGNMENT_SIZE);
		if (len < (JPEG_MARKER_LEN + JPEG_SIZE_LEN) && len != 0) {
			/* the number of bytes needed for alignment is less than the length of the marker itself, increase the number of stuffing bytes */
			len += ALIGNMENT_SIZE;
		}
		dev_dbg(dev, "total number of stuffing bytes in APP15 marker: %u\n", len);
		app15[3] = len - JPEG_MARKER_LEN;
		vectcpy(cbuff, app15, len);
		/* copy JPEG header */
		len = chunks[paged_CHUNK_HEADER].iov_len;
		dev_dbg(dev, "copy %u bytes from HEADER to common buffer\n", len);
		vectcpy(cbuff, chunks[paged_CHUNK_HEADER].iov_base, len);
		vectshrink(&chunks[paged_CHUNK_HEADER], chunks[paged_CHUNK_HEADER].iov_len);
	} else { // for Tiff (may be the same for Jpeg too - borrow data from the image data, copy up to 31 bytes to the end of the common buffer,
		// move start of the data_0 accordingly
		data_len = cbuff->iov_len;
		len = align_bytes_num(data_len, ALIGNMENT_SIZE);
		if (len){ // assuming image data > 32 bits
			vectcpy(cbuff, chunks[paged_CHUNK_DATA_0].iov_base, len);
			vectshrinkhead(&chunks[paged_CHUNK_DATA_0],len);
		}
	}
#endif

	/* check if there is enough data to continue - JPEG data length can be too short */
//	len = get_size_from      (chunks, CHUNK_DATA_0, 0, EXCLUDE_REM);
	len = get_size_from_paged(state,  CHUNK_DATA_0,    EXCLUDE_REM); // here not paged_CHUNK_DATA_0, offset will be added in get_size_from_paged

	if (len < PHY_BLOCK_SIZE) {
		size_t num = align_bytes_num(cbuff->iov_len, PHY_BLOCK_SIZE);
		dev_dbg(dev, "jpeg data is too short, delay this frame\n"); // never happened?
		if (len >= num) {
			/* there is enough data to align common buffer to sector boundary */
			if (num >= chunks[paged_CHUNK_DATA_0].iov_len) {
				vectcpy(cbuff, chunks[paged_CHUNK_DATA_0].iov_base, chunks[paged_CHUNK_DATA_0].iov_len);
				num -= chunks[paged_CHUNK_DATA_0].iov_len;
				vectshrink(&chunks[paged_CHUNK_DATA_0], chunks[paged_CHUNK_DATA_0].iov_len);
			} else {
				src = vectrpos(&chunks[paged_CHUNK_DATA_0], num);
				vectcpy(cbuff, chunks[paged_CHUNK_DATA_0].iov_base, num);
				vectshrink(&chunks[paged_CHUNK_DATA_0], num);
				num = 0;
			}
			if (chunks[paged_CHUNK_DATA_1].iov_len > 0) {
				if (num >= chunks[paged_CHUNK_DATA_1].iov_len) {
					vectcpy(cbuff, chunks[paged_CHUNK_DATA_1].iov_base, chunks[paged_CHUNK_DATA_1].iov_len);
					num -= chunks[paged_CHUNK_DATA_1].iov_len;
					vectshrink(&chunks[paged_CHUNK_DATA_1], chunks[paged_CHUNK_DATA_1].iov_len);
				} else {
					src = vectrpos(&chunks[paged_CHUNK_DATA_1], num);
					vectcpy(cbuff, chunks[paged_CHUNK_DATA_1].iov_base, num);
					vectshrink(&chunks[paged_CHUNK_DATA_1], num);
					num = 0;
				}
			}
			if (chunks[paged_CHUNK_TRAILER].iov_len > 0) {
				if (num >= chunks[paged_CHUNK_TRAILER].iov_len) {
					vectcpy(cbuff, chunks[paged_CHUNK_TRAILER].iov_base, chunks[paged_CHUNK_TRAILER].iov_len);
					num -= chunks[paged_CHUNK_TRAILER].iov_len;
					vectshrink(&chunks[paged_CHUNK_TRAILER], chunks[paged_CHUNK_TRAILER].iov_len);
				} else {
					src = vectrpos(&chunks[paged_CHUNK_TRAILER], num);
					vectcpy(cbuff, chunks[paged_CHUNK_TRAILER].iov_base, num);
					vectshrink(&chunks[paged_CHUNK_TRAILER], num);
					num = 0;
				}
			}
		} else {
			/* there is not enough data to align common buffer to sector boundary, truncate common buffer */
			data_len = cbuff->iov_len % PHY_BLOCK_SIZE;
			src = vectrpos(cbuff, data_len);
			vectcpy(chunk_rem, src, data_len);
			vectshrink(cbuff, data_len);
		}
		vectcpy(chunk_rem, chunks[paged_CHUNK_DATA_0].iov_base, chunks[paged_CHUNK_DATA_0].iov_len);
		vectshrink(&chunks[paged_CHUNK_DATA_0], chunks[paged_CHUNK_DATA_0].iov_len);
		vectcpy(chunk_rem, chunks[CHUNK_DATA_1].iov_base, chunks[paged_CHUNK_DATA_1].iov_len);
		vectshrink(&chunks[paged_CHUNK_DATA_1], chunks[CHUNK_DATA_1].iov_len);
		vectcpy(chunk_rem, chunks[CHUNK_TRAILER].iov_base, chunks[paged_CHUNK_TRAILER].iov_len);
		vectshrink(&chunks[paged_CHUNK_TRAILER], chunks[paged_CHUNK_TRAILER].iov_len);

		return;
	}

	/* align frame to sector size boundary; total size could have changed by the moment - recalculate */
//	total_sz = get_size_from      (chunks, 0, 0, INCLUDE_REM);
	total_sz = get_size_from_paged(state,  0,    INCLUDE_REM);

	len = total_sz % PHY_BLOCK_SIZE;
	dev_dbg(dev, "number of bytes crossing sector boundary: %u\n", len);
	if (len != 0) {
		if (len >= (chunks[paged_CHUNK_DATA_1].iov_len + chunks[paged_CHUNK_TRAILER].iov_len)) {
			/* current frame is not split or the second part of JPEG data is too short */
			data_len = len - chunks[paged_CHUNK_DATA_1].iov_len - chunks[paged_CHUNK_TRAILER].iov_len;
			src = vectrpos(&chunks[paged_CHUNK_DATA_0], data_len);
			vectcpy(chunk_rem, src, data_len);
			vectshrink(&chunks[paged_CHUNK_DATA_0], data_len);
			vectcpy(chunk_rem, chunks[paged_CHUNK_DATA_1].iov_base, chunks[paged_CHUNK_DATA_1].iov_len);
			vectshrink(&chunks[paged_CHUNK_DATA_1], chunks[paged_CHUNK_DATA_1].iov_len);
			vectcpy(chunk_rem, chunks[paged_CHUNK_TRAILER].iov_base, chunks[paged_CHUNK_TRAILER].iov_len);
			vectshrink(&chunks[paged_CHUNK_TRAILER], chunks[paged_CHUNK_TRAILER].iov_len);
		} else if (len >= chunks[paged_CHUNK_TRAILER].iov_len) {
			/* there is enough data in second part to align the frame */
			data_len = len - chunks[paged_CHUNK_TRAILER].iov_len;
			src = vectrpos(&chunks[paged_CHUNK_DATA_1], data_len);
			vectcpy(chunk_rem, src, data_len);
			vectshrink(&chunks[paged_CHUNK_DATA_1], data_len);
			vectcpy(chunk_rem, chunks[paged_CHUNK_TRAILER].iov_base, chunks[paged_CHUNK_TRAILER].iov_len);
			vectshrink(&chunks[paged_CHUNK_TRAILER], chunks[paged_CHUNK_TRAILER].iov_len);
		} else {
			/* the trailing marker is split by sector boundary, copy (PHY_BLOCK_SIZE - 1) bytes from
			 * JPEG data block(s) to remainder buffer and then add trailing marker */
			data_len = PHY_BLOCK_SIZE - (chunks[paged_CHUNK_TRAILER].iov_len - len);
			if (data_len >= chunks[paged_CHUNK_DATA_1].iov_len) {
				size_t cut_len = data_len - chunks[paged_CHUNK_DATA_1].iov_len;
				src = vectrpos(&chunks[paged_CHUNK_DATA_0], cut_len);
				vectcpy(chunk_rem, src, cut_len);
				vectshrink(&chunks[paged_CHUNK_DATA_0], cut_len);
				vectcpy(chunk_rem, chunks[paged_CHUNK_DATA_1].iov_base, chunks[paged_CHUNK_DATA_1].iov_len);
				vectshrink(&chunks[paged_CHUNK_DATA_1], chunks[paged_CHUNK_DATA_1].iov_len);
				vectcpy(chunk_rem, chunks[paged_CHUNK_TRAILER].iov_base, chunks[paged_CHUNK_TRAILER].iov_len);
				vectshrink(&chunks[paged_CHUNK_TRAILER], chunks[paged_CHUNK_TRAILER].iov_len);
			} else {
				src = vectrpos(&chunks[paged_CHUNK_DATA_1], data_len);
				vectcpy(chunk_rem, src, data_len);
				vectshrink(&chunks[paged_CHUNK_DATA_1], data_len);
				vectcpy(&chunks[CHUNK_REM], chunks[paged_CHUNK_TRAILER].iov_base, chunks[paged_CHUNK_TRAILER].iov_len);
				vectshrink(&chunks[paged_CHUNK_TRAILER], chunks[paged_CHUNK_TRAILER].iov_len);
			}
		}
	} else {
		/* the frame is aligned to sector boundary but some buffers may be not */
		chunks[paged_CHUNK_ALIGN].iov_base = vectrpos(cbuff, 0);
		chunks[paged_CHUNK_ALIGN].iov_len = 0;
		if (chunks[paged_CHUNK_DATA_1].iov_len == 0) {
			data_len = chunks[paged_CHUNK_DATA_0].iov_len % ALIGNMENT_SIZE;
			src = vectrpos(&chunks[CHUNK_DATA_0], data_len);
			vectcpy(&chunks[paged_CHUNK_ALIGN], src, data_len);
			vectshrink(&chunks[paged_CHUNK_DATA_0], data_len);
		} else {
			data_len = chunks[paged_CHUNK_DATA_1].iov_len % ALIGNMENT_SIZE;
			src = vectrpos(&chunks[paged_CHUNK_DATA_1], data_len);
			vectcpy(&chunks[paged_CHUNK_ALIGN], src, data_len);
			vectshrink(&chunks[paged_CHUNK_DATA_1], data_len);
		}
		vectcpy(&chunks[paged_CHUNK_ALIGN], chunks[paged_CHUNK_TRAILER].iov_base, chunks[paged_CHUNK_TRAILER].iov_len);
		vectshrink(&chunks[paged_CHUNK_TRAILER], chunks[paged_CHUNK_TRAILER].iov_len);
	}

	/* debug sanity check, should not happen */
	if (cbuff->iov_len >= COMMON_BUFF_SZ) {
		dev_dbg(NULL, "ERROR: the number of bytes copied to common buffer exceeds its size\n");
	}
	check_chunks(chunks, page);
	D6(fprintf(debug_file, "align_frame() DONE @ %06d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));

}

/** Discard buffer pointers making the command slot marked as empty */
//void reset_chunks(struct iovec *vects, int all, int page)
void reset_chunks(camogm_state *state, int all, int page)
{
	int i;
	struct iovec *vects = state->writer_params.data_chunks;
	int chunk_offset = page * FILE_CHUNKS_NUM;
	for (i = chunk_offset; i < (chunk_offset + FILE_CHUNKS_NUM); i++) {
		vects[i].iov_len = 0;
	}
	if (all) {
//		vects[CHUNK_REM].iov_len = 0;
		state->writer_params.was_CHUNK_REM.iov_len = 0;
	}
}

/** Calculate and update LBA offsets, do not count remainder buffer. Return 1 if file position should be reset to the start */
// WRONG calculation - should be split?
/*
int update_lba(camogm_state *state)
{
	int ret = 0;
	size_t total_sz;
	struct iovec *chunks = state->writer_params.data_chunks;

	total_sz = get_blocks_num(chunks, MAX_DATA_CHUNKS - 1);
	if (state->writer_params.lba_current + total_sz <= state->writer_params.lba_end) {
		state->writer_params.lba_current += total_sz;
	} else {
		state->writer_params.lba_current = state->writer_params.lba_start + total_sz;
		ret = 1;
	}

	return ret;
}
*/
/** Go through all data buffers and pick only mapped ones excluding remainder buffer */
int get_data_buffers(camogm_state *state, struct iovec *mapped, size_t mapped_sz, int page)
{
	int ret = 0;
	struct iovec *all = state->writer_params.data_chunks;
    int chunk_offset = page * FILE_CHUNKS_NUM;
	if (mapped_sz <= 0)
		return ret;

	for (int i = chunk_offset, j = 0; i < (chunk_offset + FILE_CHUNKS_NUM); i++) {
//		if (i != CHUNK_REM && all[i].iov_len != 0) {
		if (all[i].iov_len != 0) {
			if (j < mapped_sz) {
				mapped[j++] = all[i];
				ret = j;
			} else {
				ret = -1;
				break;
			}
		}
	}
	return ret;
}

/** Prepare the last remaining block of data for recording, return the number of bytes ready for recording */
int prep_last_block(camogm_state *state, int page)
{
	int ret = 0;
	size_t stuff_len;
	unsigned char *src;
	int chunk_offset = page * FILE_CHUNKS_NUM;
	struct iovec *cvect = &state->writer_params.data_chunks[chunk_offset + CHUNK_COMMON];
	struct iovec *rvect = &state->writer_params.was_CHUNK_REM;

	if (rvect->iov_len != 0) {
		stuff_len = PHY_BLOCK_SIZE - rvect->iov_len;
		src = vectrpos(rvect, 0);
		memset(src, 0, stuff_len);
		rvect->iov_len += stuff_len;
		ret = rvect->iov_len;
		vectcpy(cvect, rvect->iov_base, rvect->iov_len);
		vectshrink(rvect, rvect->iov_len);
	}

	return ret;
}

/** Convert LBA to byte offset used for lseek */
off64_t lba_to_offset(uint64_t lba)
//uint64_t lba_to_offset(uint64_t lba)
{
	return lba * PHY_BLOCK_SIZE;
}
