/** @file camogm_jpeg.c
 * @brief Provides writing to series of individual JPEG files for @e camogm
 * @copyright Copyright (C) 2016 Elphel, Inc.
 *
 * @par <b>License</b>
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/** @brief This define is needed to use lseek64 and should be set before includes */
#define _LARGEFILE64_SOURCE

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <time.h>
#include <elphel/c313a.h>
#include <elphel/ahci_cmd.h>

#include "camogm_jpeg.h"
#include "camogm_read.h"
#include "camogm_align.h"


/** State file record format. It includes device path in /dev, starting, current and ending LBAs */
#define STATE_FILE_FORMAT         "%s\t%llu\t%llu\t%llu\n"
/** Status file update period, in seconds */
#define STAT_UPDATE_PERIOD        60

/* forward declarations */
static void *jpeg_writer(void *thread_args);
static int save_state_file(const camogm_state *state);

/** Get starting and endign LBAs of the partition specified as raw device buffer */
static int get_disk_range(struct range *range)
{
	FILE *f;

	// get raw device buffer starting postion on disk
	f = fopen(SYSFS_AHCI_LBA_START, "r");
	if (f == NULL) {
		return -1;
	}
	fscanf(f, "%llu\n", &range->from);
	fclose(f);

	// get raw device buffer ending postion on disk
	f = fopen(SYSFS_AHCI_LBA_END, "r");
	if (f == NULL) {
		return -1;
	}
	fscanf(f, "%llu\n", &range->to);
	fclose(f);

	return 0;
}

/** Get write pointer from a file. This functions check not only the name of a partition, but
 * its geometry as well */
static int find_state(FILE *f, uint64_t *pos, const rawdev_buffer *rawdev)
{
	size_t len;
	uint64_t start_pos, curr_pos, end_pos;
	struct range range;
	char buff[ELPHEL_PATH_MAX];
	char dev_name[ELPHEL_PATH_MAX];

	if (f == NULL || pos == NULL)
		return -1;
	if (get_disk_range(&range) != 0) {
		return -1;
	}

	// skip first line containing file header
	fgets(buff, ELPHEL_PATH_MAX, f);
	while (fgets(buff, ELPHEL_PATH_MAX, f) != NULL) {
		sscanf(buff, STATE_FILE_FORMAT, dev_name, &start_pos, &curr_pos, &end_pos);
		len = strlen(dev_name);
		if (strncmp(rawdev->rawdev_path, dev_name, len) == 0 &&
				range.from == start_pos &&
				range.to == end_pos) {
			*pos = curr_pos;
			break;
		}
	}

	return 0;
}

/** Read state from file and restore disk write pointer */
int open_state_file(const rawdev_buffer *rawdev, uint64_t *current_pos)
{
	FILE *f;
	int ret = 0;
	uint64_t lba_pos;

	if (strlen(rawdev->state_path) == 0) {
		return ret;
	}

	f = fopen(rawdev->state_path, "r");
	if (f != NULL) {
		if (find_state(f, &lba_pos, rawdev) != -1) {
			*current_pos = lba_pos;
			D0(fprintf(debug_file, "Got starting LBA from state file: %llu\n", lba_pos));
		}
		fclose(f);
	} else {
		ret = -1;
	}

	return ret;
}

/** Save current position of the disk write pointer */
static int save_state_file(const camogm_state *state)
{
	int ret = 0;
	FILE *f;
	const rawdev_buffer *rawdev = &state->rawdev;
	const struct writer_params *params = &state->writer_params;

	if (strlen(rawdev->state_path) == 0) {
		return ret;
	}

	// save pointers to a regular file
	f = fopen(rawdev->state_path, "w");
	if (f == NULL) {
		return -1;
	}
	fprintf(f, "Device\t\tStart LBA\tCurrent LBA\tEnd LBA\n");
	fprintf(f, STATE_FILE_FORMAT, rawdev->rawdev_path, params->lba_start, params->lba_current, params->lba_end);
	fflush(f);
	fsync(fileno(f));
	fclose(f);

	return ret;
}

/**
 * @brief Initialize synchronization resources for disk writing thread and then start this thread. This function
 * is call each time JPEG format is set or changed, thus we need to check the state of writing thread before
 * initialization to prevent spawning multiple threads.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      0 if initialization was successful and negative value otherwise
 */
int camogm_init_jpeg(camogm_state *state)
{
	int ret = 0;
	int ret_val;

	if (state->writer_params.state == STATE_STOPPED) {
		ret_val = pthread_cond_init(&state->writer_params.main_cond, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize conditional variable for main thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = pthread_cond_init(&state->writer_params.writer_cond, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize conditional variable for writing thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = pthread_mutex_init(&state->writer_params.writer_mutex, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize mutex for writing thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = pthread_create(&state->writer_params.writer_thread, NULL, jpeg_writer, (void *)state);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not start writer thread: %s\n", strerror(ret_val)));
			ret = -1;
		}

		ret_val = init_align_buffers(state);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize alignment buffers\n"));
			ret = -1;
		}
	}

	return ret;
}

/**
 * @brief Stop disk writing thread and free its resources. This function is called after the program receives 'exit' command.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      None
 */
void camogm_free_jpeg(camogm_state *state)
{
	pthread_cond_destroy(&state->writer_params.main_cond);
	pthread_cond_destroy(&state->writer_params.writer_cond);
	pthread_mutex_destroy(&state->writer_params.writer_mutex);

	// terminate writing thread
	pthread_mutex_lock(&state->writer_params.writer_mutex);
	state->writer_params.exit_thread = true;
	pthread_cond_signal(&state->writer_params.writer_cond);
	pthread_mutex_unlock(&state->writer_params.writer_mutex);
	pthread_join(state->writer_params.writer_thread, NULL);
	state->writer_params.exit_thread = false;

	deinit_align_buffers(state);
}

/** Calculate the total length of current frame */
int64_t camogm_get_jpeg_size(camogm_state *state)
{
	int64_t len = 0;

	for (int i = 0; i < state->chunk_index - 1; i++) {
		len += state->packetchunks[i + 1].bytes;
	}

	return len;
}

/**
 * @brief Called every time the JPEG files recording is started.
 *
 * This function checks if the raw device write is initiated and tries to open the device specified. The device
 * will be closed in #camogm_end_jpeg function.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      0 if the device was opened successfully and negative error code otherwise
 */
int camogm_start_jpeg(camogm_state *state)
{
	char * slash;
	int rslt;
	off64_t offset;

	if (!state->rawdev_op) {
		strcpy(state->path, state->path_prefix);   // make state->path a directory name (will be replaced when the frames will be written)
		slash = strrchr(state->path, '/');
		D2(fprintf(debug_file, "camogm_start_jpeg\n"));
		if (slash) {
			D3(fprintf(debug_file, "Full path %s\n", state->path));
			slash[0] = '\0';                       // truncate path to the directory name
			D3(fprintf(debug_file, "directory path %s\n", state->path));
			rslt = mkdir(state->path, 0777);
			D3(fprintf(debug_file, "mkdir (%s, 0777) returned %d, errno=%d\n", state->path, rslt, errno));
			if ((rslt < 0) && (errno != EEXIST)) { // already exists is OK
				D0(fprintf(debug_file, "Error creating directory %s, errno=%d\n", state->path, errno));
				return -CAMOGM_FRAME_FILE_ERR;
			}
		}
	} else {
		if (open_state_file(&state->rawdev, &state->writer_params.lba_current) != 0) {
			D0(fprintf(debug_file, "Could not get write pointer from state file, recording will start from the beginning of partition: "
					"%s\n", state->rawdev.rawdev_path));
		}
		state->writer_params.blockdev_fd = open(state->rawdev.rawdev_path, O_RDWR);
		if (state->writer_params.blockdev_fd < 0) {
			D0(fprintf(debug_file, "Error opening block device: %s\n", state->rawdev.rawdev_path));
			return -CAMOGM_FRAME_FILE_ERR;
		}
		offset = lba_to_offset(state->writer_params.lba_current - state->writer_params.lba_start);
		lseek64(state->writer_params.blockdev_fd, offset, SEEK_SET);
		D3(fprintf(debug_file, "Open block device: %s, offset in bytes: %llu\n", state->rawdev.rawdev_path, offset));
		state->writer_params.stat_update = time(NULL);
	}

	return 0;
}

/**
 * @brief Write single JPEG frame
 *
 * This function will write single JPEG file
 * @param   state   a pointer to a structure containing current state
 * @return
 */
int camogm_frame_jpeg(camogm_state *state)
{
	int i, j;
	ssize_t iovlen, l = 0;
	struct iovec chunks_iovec[8];
	int port = state->port_num;
	time_t curr_time;
	int is_tiff = state->this_frame_params[port].color == COLORMODE_RAW;
    if (is_tiff) {
    	sprintf(state->path, "%s%d_%010ld_%06ld.tiff", state->path_prefix, port, state->this_frame_params[port].timestamp_sec, state->this_frame_params[port].timestamp_usec);
    } else {
    	sprintf(state->path, "%s%d_%010ld_%06ld.jpeg", state->path_prefix, port, state->this_frame_params[port].timestamp_sec, state->this_frame_params[port].timestamp_usec);
    }
	if (!state->rawdev_op) {
		l = 0;
		for (i = 0; i < (state->chunk_index) - 1; i++) {
			chunks_iovec[i].iov_base = state->packetchunks[i + 1].chunk;
			chunks_iovec[i].iov_len = state->packetchunks[i + 1].bytes;
			l += chunks_iovec[i].iov_len;
		}
		if (((state->ivf = open(state->path, O_RDWR | O_CREAT, 0777))) < 0) {
			D0(fprintf(debug_file, "Error opening %s for writing, returned %d, errno=%d\n", state->path, state->ivf, errno));
			return -CAMOGM_FRAME_FILE_ERR;
		}
		iovlen = writev(state->ivf, chunks_iovec, (state->chunk_index) - 1);
		if (iovlen < l) {
			j = errno;
			D0(fprintf(debug_file, "writev error %d (returned %d, expected %d)\n", j, iovlen, l));
			close(state->ivf);
			return -CAMOGM_FRAME_FILE_ERR;
		}
		state->rawdev.last_jpeg_size = l;
		close(state->ivf);
	} else {
		D6(fprintf(debug_file, "\ndump iovect array for port %u\n", state->port_num));
		for (int i = 0; i < state->chunk_index - 1; i++) {
			D6(fprintf(debug_file, "ptr: %p, length: %ld\n", state->packetchunks[i + 1].chunk, state->packetchunks[i + 1].bytes));
		}
		D6(fprintf(debug_file, "waiting for empty buffer @ %06d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
		// next frame is ready for recording, signal this to the writer thread
		pthread_mutex_lock(&state->writer_params.writer_mutex);
//		while (state->writer_params.data_ready)
//			pthread_cond_wait(&state->writer_params.main_cond, &state->writer_params.writer_mutex);
		// do not run ahead by >= FILE_CHUNKS_PAGES (0 or 1 for just FILE_CHUNKS_PAGES == 2)
		while ((state->writer_params.chunk_page_prep - state->writer_params.chunk_page_write + (2 * FILE_CHUNKS_PAGES)) % (2 * FILE_CHUNKS_PAGES) >= FILE_CHUNKS_PAGES)
			pthread_cond_wait(&state->writer_params.main_cond, &state->writer_params.writer_mutex);
		pthread_mutex_unlock(&state->writer_params.writer_mutex);
		D6(fprintf(debug_file, "_13a_: got empty buffer @ %06d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
		align_frame(state);
		/* // Moved to write thread
		if (update_lba(state) == 1) {
			D0(fprintf(debug_file, "The end of block device reached, continue recording from start\n"));
			lseek(state->writer_params.blockdev_fd, 0, SEEK_SET);
		}
		D6(fprintf(debug_file, "Block device positions: start = %llu, current = %llu, end = %llu\n",
				state->writer_params.lba_start, state->writer_params.lba_current, state->writer_params.lba_end));
       */
		// proceed if last frame was recorded without errors
		D6(fprintf(debug_file, "_13a1_:frame aligned @ %06d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
		pthread_mutex_lock(&state->writer_params.writer_mutex);
		if (state->writer_params.last_ret_val == 0) {
			state->writer_params.chunk_page_prep = (state->writer_params.chunk_page_prep + 1) % (2 * FILE_CHUNKS_PAGES);
//			state->writer_params.data_ready = true;
			pthread_cond_signal(&state->writer_params.writer_cond);
			D6(fprintf(debug_file, "_13b_: chunk_page_prep = %d (%d), signal writer_cond @ %06d\n",state->writer_params.chunk_page_prep, state->writer_params.chunk_page_write, get_fpga_usec(state->fd_fparmsall[0], 0)));
		}
		pthread_mutex_unlock(&state->writer_params.writer_mutex);
		if (state->writer_params.last_ret_val != 0) {
			return state->writer_params.last_ret_val;
		}
		// update status file if time has come
		curr_time = time(NULL);
		if (difftime(curr_time, state->writer_params.stat_update) > STAT_UPDATE_PERIOD) {
			save_state_file(state);
			D6(fprintf(debug_file, "_13c_: saved state file @ %06d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
		}
	}

	return 0;
}

/**
 * @brief Finish JPEG file write operation
 *
 * This function checks whether raw device write was on and closes raw device file.
 * @param   state   a pointer to a structure containing current state
 * @return  0 if the device was closed successfully and -1 otherwise
 */
int camogm_end_jpeg(camogm_state *state) // TODO: Resolve multippage !
{
	int ret = 0;
	int bytes;
	ssize_t iovlen;

	if (state->rawdev_op) {
		// write any remaining data, do not use writer thread as there can be only one block left CHUNK_REM buffer
		pthread_mutex_lock(&state->writer_params.writer_mutex);

//		while (state->writer_params.data_ready)
		while (state->writer_params.chunk_page_prep != state->writer_params.chunk_page_write)
			// wait for previous frame to be recorded first if it has not be recorded by the moment
			pthread_cond_wait(&state->writer_params.main_cond, &state->writer_params.writer_mutex);

		bytes = prep_last_block(state, state->writer_params.chunk_page_prep  % FILE_CHUNKS_PAGES);
		if (bytes > 0) {
			D6(fprintf(debug_file, "Write last block of data, size = %d\n", bytes));
			// the remaining data block is placed in CHUNK_COMMON buffer, write just this buffer
			iovlen = writev(state->writer_params.blockdev_fd,
					&state->writer_params.data_chunks[CHUNK_COMMON + (state->writer_params.chunk_page_prep % FILE_CHUNKS_PAGES) * FILE_CHUNKS_NUM], 1);
			if (iovlen < bytes) {
				D0(fprintf(debug_file, "writev error: %s (returned %i, expected %i)\n", strerror(errno), iovlen, bytes));
				state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
			} else {
				// update statistic, just one block written
				state->writer_params.lba_current += 1;
				state->rawdev.total_rec_len += bytes;
			}
//			reset_chunks(state->writer_params.data_chunks, 1);
			 //*************** See if another buffer(s) must be cleared too? or even use chunk_page_write
			reset_chunks(state, 1, state->writer_params.chunk_page_prep % FILE_CHUNKS_PAGES);
		}
		pthread_mutex_unlock(&state->writer_params.writer_mutex);

		D6(fprintf(debug_file, "Closing block device %s\n", state->rawdev.rawdev_path));
		ret = close(state->writer_params.blockdev_fd);
		if (ret == -1)
			D0(fprintf(debug_file, "Error: %s\n", strerror(errno)));

		save_state_file(state);
	}
	return ret;
}

/**
 * @brief Disk writing thread. This thread holds local copy of a structure containing current state
 * of the program and updates it on a signal from main thread every time new frame is ready for recording.
 * @param[in]   thread_args   a pointer to a structure containing current state
 * @return      None
 */
void *jpeg_writer(void *thread_args)
{
	int dbg_us1, dbg_us2, dbg_us; // save usec for difference
	int reset_rem;
	int chunk_index, i0, i1;
	ssize_t iovlen, l, ll, l0, l1, blocks_write, wpage;
	bool process = true, has_data_to_write = false;
	struct iovec chunks_iovec[FILE_CHUNKS_PAGES][FILE_CHUNKS_NUM];
	struct iovec split_chunk;

	camogm_state *state = (camogm_state *)thread_args;
	struct writer_params *params = &state->writer_params;
	unsigned char dummy_buff[PHY_BLOCK_SIZE];

	memset((void *)chunks_iovec, 0, sizeof(struct iovec) * FILE_CHUNKS_PAGES * FILE_CHUNKS_NUM);
//	pthread_mutex_lock(&params->writer_mutex);
	params->state = STATE_RUNNING;
	D6(fprintf(debug_file, "_w000_:  Before jpeg_writer loop @ %06d\n", get_fpga_usec(state->fd_fparmsall[0], 0))); // before the debug file is set !
	while (process) {
		/*
		while (!params->data_ready && !params->exit_thread) {
			pthread_cond_wait(&params->writer_cond, &params->writer_mutex);
		} // get rid of params->data_ready
		*/
		pthread_mutex_lock(&params->writer_mutex);  // first time before the debug file is set !
		D6(fprintf(debug_file, "_w00_: (%d, %d) @ %06d\n", params->chunk_page_prep, params->chunk_page_write, get_fpga_usec(state->fd_fparmsall[0], 0)));
		while ((params->chunk_page_prep == params->chunk_page_write) && !params->exit_thread) {
			pthread_cond_wait(&params->writer_cond, &params->writer_mutex);
		}

		if (params->exit_thread) {
			process = false;
		}
		has_data_to_write = params->chunk_page_prep != params->chunk_page_write;
		pthread_mutex_unlock(&params->writer_mutex);
		D6(fprintf(debug_file, "_w00.5_: (%d, %d) @ %06d\n", params->chunk_page_prep, params->chunk_page_write,get_fpga_usec(state->fd_fparmsall[0], 0)));
//		if (params->data_ready) {
		if (has_data_to_write) {
			D6(fprintf(debug_file, "_w01_: got write data (%d > %d) @ %06d\n", params->chunk_page_prep, params->chunk_page_write,get_fpga_usec(state->fd_fparmsall[0], 0)));
			/* dummy read cycle from (approximately) the beginning of previous frame;
			 * this is a debug feature used to find disk errors */
			/*//temporary disable for simplicity 05/21/2023
			if (state->writer_params.dummy_read) {
				ssize_t data_len;
				off64_t curr_offset = lseek64(state->writer_params.blockdev_fd, 0, SEEK_CUR);
				off64_t offset = lba_to_offset(state->writer_params.lba_current - state->writer_params.lba_start) - state->rawdev.last_jpeg_size;
				offset = offset / PHY_BLOCK_SIZE;
				lseek64(state->writer_params.blockdev_fd, offset, SEEK_SET);
				data_len = read(state->writer_params.blockdev_fd, dummy_buff, PHY_BLOCK_SIZE);
				if (data_len < PHY_BLOCK_SIZE) {
					D6(fprintf(debug_file, "Dummy read error: requested %d, read %d, %s\n", PHY_BLOCK_SIZE, data_len, strerror(errno)));
				}
				lseek64(state->writer_params.blockdev_fd, curr_offset, SEEK_SET);
			}
			*/
			/* end of dummy read cycle */

			l = 0;
			reset_rem = 0;
			state->writer_params.last_ret_val = 0;
			 // inputs from state->writer_params.data_chunks, returns number of result chunks
			wpage = state->writer_params.chunk_page_write % FILE_CHUNKS_PAGES;
			chunk_index = get_data_buffers(state, chunks_iovec[wpage], FILE_CHUNKS_NUM, wpage);
			// assuming write is block-aligned
			if (chunk_index > 0) {
				for (int i = 0; i < chunk_index; i++)
					l += chunks_iovec[wpage][i].iov_len;
				// Moved rollover here
				blocks_write = l / PHY_BLOCK_SIZE;
				if (state->writer_params.lba_current + blocks_write <= state->writer_params.lba_end) { // almost always - single write
					state->writer_params.lba_current += blocks_write;
//					D6(fprintf(debug_file, "_w02_: starting writev @ %06d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
					D6(dbg_us1=get_fpga_usec(state->fd_fparmsall[0], 0); fprintf(debug_file, "_w02_: starting writev @ %06d\n",dbg_us1));
					iovlen = writev(state->writer_params.blockdev_fd, chunks_iovec[wpage], chunk_index);
//					D6(fprintf(debug_file, "_w03_: finished writev @ %06d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
					D6(dbg_us=get_fpga_usec(state->fd_fparmsall[0], 0); fprintf(debug_file, "_w03_: finished writev @ %06d (+ %06d, +%06d)\n",dbg_us, (dbg_us-dbg_us1), (dbg_us-dbg_us2)); dbg_us2=dbg_us);
					if (iovlen < l) {
						D0(fprintf(debug_file, "writev error: %s (returned %i, expected %i)\n", strerror(errno), iovlen, l));
						state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
						reset_rem = 1;
						lseek(state->writer_params.blockdev_fd,
								(state->writer_params.lba_current - state->writer_params.lba_start) * PHY_BLOCK_SIZE,
								SEEK_SET); // update as if written?
					} else {
						// update statistic
						state->rawdev.last_jpeg_size = l;
						state->rawdev.total_rec_len += state->rawdev.last_jpeg_size;
						D6(fprintf(debug_file, "Current position in block device: %lld\n", lseek64(state->writer_params.blockdev_fd, 0, SEEK_CUR)));
					}
				} else { // Split into two writes: to the end of partition and startion from the beginning
					l0 = (state->writer_params.lba_end - state->writer_params.lba_current) * PHY_BLOCK_SIZE; // remaining to be written before rollover
					l1 = l - l0; // remaining to be written after rollover
					i1 = 0; // start of the second write
					if (l0 > 0) { // should be always
						ll = 0;
						for (i0 = 0; i0 < chunk_index; i0++){
							if (ll + chunks_iovec[wpage][i0].iov_len > l0){
								break;
							}
							ll += chunks_iovec[wpage][i0].iov_len;
						}
						// save split chunk
						split_chunk.iov_base = chunks_iovec[wpage][i0].iov_base;
						split_chunk.iov_len =  chunks_iovec[wpage][i0].iov_len;
						i1 = i0; // index of the first chunk in a second write
						if (ll < l0) { // is it a partial chunk?
							chunks_iovec[wpage][i0].iov_len = l0 - ll;
							i0++;     // include split chunk in the first write
						}
						// write the first group (before rollover)
						state->writer_params.lba_current = state->writer_params.lba_start;
						iovlen = writev(state->writer_params.blockdev_fd, chunks_iovec[wpage], i0);
						lseek(state->writer_params.blockdev_fd, 0, SEEK_SET); // regardless of error - rewind
						if (iovlen < l0) {
							D0(fprintf(debug_file, "(1 of 2) writev error: %s (returned %i, expected %i)\n", strerror(errno), iovlen, l0));
							state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
							reset_rem = 1;
						} else {
							// do not update statistic after the first writed, do it after the second
							D6(fprintf(debug_file, "(1 of 2) Current position in block device: %lld\n", lseek64(state->writer_params.blockdev_fd, 0, SEEK_CUR)));
							// update split chunk to use its second part, shift chunks
							if (l1 > 0) { // length to be written after rollover
								if (chunks_iovec[wpage][i1].iov_len != split_chunk.iov_len) { // it was split
									D0(fprintf(debug_file, "chunk %d was split, full length = %d, writtren = %d\n", i1, split_chunk.iov_len, chunks_iovec[wpage][i1].iov_len));
									chunks_iovec[wpage][i1].iov_base += chunks_iovec[wpage][i1].iov_len;
									chunks_iovec[wpage][i1].iov_len = split_chunk.iov_len - chunks_iovec[wpage][i1].iov_len;
								}
								// shift chunks
								for (i0 = i1; i0 < chunk_index; i0++){
									chunks_iovec[wpage][i0 - i1] = chunks_iovec[wpage][i0];
								}
								iovlen = writev(state->writer_params.blockdev_fd, chunks_iovec[wpage], chunk_index - i1);
								state->writer_params.lba_current = state->writer_params.lba_start + l1 / PHY_BLOCK_SIZE;
								if (iovlen < l1) {
									D0(fprintf(debug_file, "(2 of 2) writev error: %s (returned %i, expected %i)\n", strerror(errno), iovlen, l1));
									state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
									reset_rem = 1;
								} else {
									// update statistic (for both parts of the split write)
									state->rawdev.last_jpeg_size = l;
									state->rawdev.total_rec_len += state->rawdev.last_jpeg_size;
									D6(fprintf(debug_file, "(1 of 2) Current position in block device: %lld\n", lseek64(state->writer_params.blockdev_fd, 0, SEEK_CUR)));
								}
							}
						}
					}
				}
				D6(fprintf(debug_file, "_w04: wrote data @ %06d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
				D6(fprintf(debug_file, "_w05_: Block device positions: start = %llu, current = %llu, end = %llu\n",
						state->writer_params.lba_start, state->writer_params.lba_current, state->writer_params.lba_end));
			} else {
				D0(fprintf(debug_file, "data vector mapping error: %d)\n", chunk_index));
				state->writer_params.last_ret_val = -CAMOGM_FRAME_FILE_ERR;
				reset_rem = 1;
			}

			// release main thread
//			reset_chunks(state->writer_params.data_chunks, reset_rem);
			reset_chunks(state, reset_rem, state->writer_params.chunk_page_write % FILE_CHUNKS_PAGES);

			pthread_mutex_lock(&params->writer_mutex);
			params->chunk_page_write = (params->chunk_page_write + 1) % (2 * FILE_CHUNKS_PAGES);
//			params->data_ready = false;
			if (params->chunk_page_prep == params->chunk_page_write) {
				pthread_cond_signal(&params->main_cond); // **********  Do not signal if can write more !
				D6(fprintf(debug_file, "_w06_: wrote data (%d , %d), buffer empty, signal main thread @ %06d\n", params->chunk_page_prep, params->chunk_page_write,get_fpga_usec(state->fd_fparmsall[0], 0)));
			} else {
				D6(fprintf(debug_file, "_w06a_: wrote data (%d , %d), has more to write, do not signal! @ %06d\n", params->chunk_page_prep, params->chunk_page_write,get_fpga_usec(state->fd_fparmsall[0], 0)));

			}
			pthread_mutex_unlock(&params->writer_mutex); // do not release untill waiting for writev() or waiting for data to write
		}
	}
	params->state = STATE_STOPPED;
//	pthread_mutex_unlock(&state->writer_params.writer_mutex);
	D5(fprintf(debug_file, "_w07_: Exit from recording thread\n")); // never?

	return NULL;
}
